package net.javainthebox.annotation;

import java.io.IOException;
import java.io.PrintStream;
import java.util.Set;
import java.util.List;
import javax.annotation.processing.AbstractProcessor;
import javax.annotation.processing.Filer;
import javax.annotation.processing.Messager;
import javax.annotation.processing.RoundEnvironment;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.ElementVisitor;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.Name;
import javax.lang.model.element.TypeElement;
import javax.lang.model.element.TypeParameterElement;
import javax.lang.model.element.VariableElement;
import javax.lang.model.type.TypeMirror;
import javax.lang.model.util.ElementScanner6;
import javax.tools.Diagnostic;
import javax.tools.JavaFileObject;

@SupportedSourceVersion(SourceVersion.RELEASE_6)
@SupportedAnnotationTypes({
    "net.javainthebox.annotation.Abstraction",
    "net.javainthebox.annotation.AbstractionMethod"
})
public class AbstractionProcessor extends AbstractProcessor {
    private Messager messager;
    private Filer filer;
    private PrintStream stream;

    class AbstractionScanner 
        extends ElementScanner6<Void, Void> {

        @Override
        public Void visitType(TypeElement element, Void p) {
            if (element.getAnnotation(Abstraction.class) != null) {
                // AbstractionAme[VŏCĂꍇ
                // C^tF[X̒`o
                try {
                    makeInterfaceDefinition(element);
                } catch (IOException ex) {
                    messager.printMessage(Diagnostic.Kind.ERROR, 
                                          ex.getMessage(), 
                                          element);
                }
            }

            return super.visitType(element, p);
        }

        @Override
        public Void visitExecutable(ExecutableElement element, Void p) {
            if (element.getAnnotation(AbstractionMethod.class) != null) {
                // AbstractionMethodAme[VŏCĂꍇ
                // C^tF[XŃ\bh`s
                makeMethodDefinition(element);
            }
            
            return super.visitExecutable(element, p);
        }

        // C^tF[X̒`o
        private void makeInterfaceDefinition(TypeElement element)
                                             throws IOException {
            // C^tF[X̖O߂
            // ł͌̃NX̃pbP[W͖A
            // ftHgpbP[WɃC^tF[X쐬
            // C^tF[X I + ̖O
            Name name = element.getQualifiedName();
            int index = name.toString().lastIndexOf('.');
            String stubName = "I" + name.subSequence(index + 1, name.length());
            
            // C^tF[X̃\[Xt@C쐬
            JavaFileObject stubFile = filer.createSourceFile(stubName);
            // Xg[擾
            stream = new PrintStream(stubFile.openOutputStream());
            
            // ꂽƂ Generated Ame[Vt
            stream.println("@javax.annotation.Generated(\"StubProcessor\")");
            
            // C^tF[X`̏o
            stream.println("public interface " + stubName + " {");
        }
        
        // \bh`̏o
        private void makeMethodDefinition(ExecutableElement element) {
            stream.print("    public " 
                         + element.getReturnType() + " " 
                         + element.getSimpleName() + "(");
            
            // \bh̉擾
            List<? extends VariableElement> args 
                = element.getParameters();
            
            if (args.size() != 0) {
                // 񋓂
                int i = 0;
                for (; i < args.size() - 1; i++) {
                    stream.print(args.get(i).asType() + " " 
                                 + args.get(i) + ", ");
                }
                stream.print(args.get(i).asType() + " " 
                             + args.get(i) + ")");
            } else {
                // ȂΊʂ
                stream.print(")");
            }
            
            // throwsO擾
            List<? extends TypeMirror> exceptions
                = element.getThrownTypes();
            
            if (exceptions.size() != 0) {
                // O
                stream.print(" throws ");
                
                int i = 0;
                for (; i < exceptions.size() - 1; i++) {
                    stream.print(exceptions.get(i) + ", ");
                }
                stream.print(exceptions.get(i) + ";");
            } else {
                // OȂ΃Z~RōsI点
                stream.println(";");
            }
        }
    }

    @Override
    public boolean process(
            Set<? extends TypeElement> annotations, 
            RoundEnvironment roundEnvironment) {

        // FilerIuWFNg̎擾
        filer = processingEnv.getFiler();
        // MessagerIuWFNg̎擾
        messager = processingEnv.getMessager();

        // XLi̐
        AbstractionScanner scanner = new AbstractionScanner();
 
        for (Element root: roundEnvironment.getRootElements()) {
            // [gvfXLĂ
            scanner.scan(root);

            if (stream != null) {
                try {
                    // C^tF[X̍Ō̕o
                    makeInterfaceFooter();
                } catch (IOException ex) {
                    messager.printMessage(Diagnostic.Kind.ERROR, 
                                          ex.getMessage(), 
                                          root);
                }
            }
        }

        return true;
    }

    // NX̍Ō̊ʂo
    private void makeInterfaceFooter() throws IOException {
        stream.println("}");
        stream.close();
        stream = null;
    }
}
